/* mod_urstate - ncaptool plugin implementing the urstate heuristic */

/*
 * Copyright (c) 2008 by Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <arpa/nameser.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <Judy.h>

#define ISC_CHECK_NONE 1
#include <isc/list.h>

#include <ncaptool.h>

/* Types. */

struct qrkey {
	uint32_t		ip[2];
	uint16_t		port[2];
	uint16_t		qid;
	uint16_t		qclass;
	uint8_t			qtype;
	char			qname[0];
} __attribute__((packed));

struct qrval {
	signed			count;
	struct ncap_msg *	msgq;
	struct ncap_msg *	msgr;
};

struct qrent {
	ISC_LINK(struct qrent)	link;
	int			len;
	struct qrkey		key;
};

/* Macros. */

#define SWAP(a, b) do { \
	char c[sizeof(a)]; \
	memcpy((void *)&c, (void *)&a, sizeof(c)); \
	memcpy((void *)&a, (void *)&b, sizeof(a)); \
	memcpy((void *)&b, (void *)&c, sizeof(b)); \
} while (0)

#ifndef NCAP_OK
#define NCAP_OK(n,e,a) do { \
	if (n->e != ncap_success) { \
		fprintf(stderr, #e ": %s\n", n->errstr); \
		a; \
	} \
} while (0)
#endif

/* Forward declarations for functions exported via module context. */

static int fini(void);
static int init(struct ncaptool_ctx *, int, char **);
static int proc(ncap_t, void *, ncap_msg_ct, const char *);
static int sinks_enab(void);
static int sinks_open(void);
static void drainsinks(void);

/* Forward declarations for private module functions. */

static int dns_makename(const char *, char *);
static int process_msg(ncap_t, ncap_msg_ct);
static int same_msg(ncap_msg_ct, ncap_msg_ct);
static struct ncap_msg *dup_msg(struct ncap_msg *);
static void destroy(void);
static void dump_msg(ncap_t, ncap_msg_ct);
static void free_msg(struct ncap_msg *);
static void trim_entries(void);
static void usage(const char *msg);

/* Module globals. */

static struct ncaptool_ctx *	ctx;
static void *			urstate;
static ISC_LIST(struct qrent)	ent_list;
static struct ncaptool_fpsink	urf;

static time_t			start_time;
static unsigned			suppress_duration;

static uint64_t			count_dup_query;
static uint64_t			count_dup_response;
static uint64_t			count_entries;
static uint64_t			count_entries_limit;
static uint64_t			count_messages_out;
static uint64_t			count_messages_proc;

static int			clear_id;
static int			clear_ports;
static int			pass_through;
static int			verbose;

#define ENTRIES_LIMIT		1048576
#define SUPPRESS_DURATION	30
#define RESPONSE_THRESHOLD	0

/* Module context. */

struct ncaptool_msgmod ncaptool_msgmod_ctx =
{
	.mver = NCAPTOOL_MSGMOD_VERSION,
	.name = "mod_urstate",
	.init = &init,
	.fini = &fini,
	.proc = &proc,
	.sinks_open = &sinks_open,
	.sinks_enab = &sinks_enab,
	.drainsinks = &drainsinks
};

/* Functions. */

static int
init(struct ncaptool_ctx *my_ctx, int argc, char **argv) {
	int ch;
	char *t;

	ctx = my_ctx;
	count_entries_limit = ENTRIES_LIMIT;
	suppress_duration = SUPPRESS_DURATION;

	ISC_LIST_INIT(ent_list);

	while ((ch = getopt(argc, argv, "o:t:L:DPv")) != -1) {
		switch (ch) {
		case 'o':
			if (!ctx->getsink(optarg, &urf))
				usage("bad -o sink");
			(ctx->sinks)++;
			pass_through = 1;
			break;
		case 't':
			suppress_duration = strtoul(optarg, &t, 0);
			if (*t != '\0') {
				usage("bad -t number");
				return (-1);
			}
			break;
		case 'L':
			count_entries_limit = strtoul(optarg, &t, 0);
			if (*t != '\0') {
				usage("bad -L number");
				return (-1);
			}
			break;
		case 'D':
			clear_id = 1;
			break;
		case 'P':
			clear_ports = 1;
			break;
		case 'v':
			verbose += 1;
			break;
		default:
			usage("unrecognized argument");
			return (-1);
		}
	}
	if (ctx->samesink(&ctx->trace, &urf)) {
		usage("ncaptool sink and mod_urstate sink can't be the same");
		return (-1);
	}
	return (0);
}

static int
fini(void) {
	ctx->disablesink(&urf, ctx);
	destroy();
	return (0);
}

static int
sinks_open(void) {
	return (NCAPTOOL_SINK_ACTIVE(&urf));
}

static int
sinks_enab(void) {
	return (NCAPTOOL_SINK_ENABLED(&urf));
}

static void
drainsinks(void) {
	if (NCAPTOOL_SINK_ENABLED(&urf)) {
		ctx->drainsink(&urf, ctx);
		count_messages_out = 0;
	}
}

static int
proc(ncap_t ncap,
	   void *user		__attribute__((unused)),
	   ncap_msg_ct msg,
	   const char *label	__attribute__((unused)))
{
	if (start_time == 0)
		 start_time = ctx->now.tv_sec;
	if (!(msg->np == ncap_ip4 && msg->tp == ncap_udp))
		return (0);
	if (count_entries > count_entries_limit)
		trim_entries();

	if (NCAPTOOL_SINK_ENABLED(&urf) && !NCAPTOOL_SINK_ACTIVE(&urf)) {
		if (!ctx->opensink(&urf, ctx->now,
			ctx->kicker != NULL ? "ncap" : NULL))
		{
			ctx->disablesink(&urf, ctx);
		} else {
			ctx->opened_sink = 1;
		}
	}

	return (process_msg(ncap, msg));
}

/* Private functions. */

static void
trim_entries(void) {
	struct qrent *ent;
	struct qrval **val;
	int rc;
	while (count_entries > count_entries_limit) {
		ent = ISC_LIST_HEAD(ent_list);
		JHSG(val, urstate, &ent->key, ent->len);
		if (val != NULL) {
			free_msg((*val)->msgq);
			free_msg((*val)->msgr);
			free(*val);
			JHSD(rc, urstate, &ent->key, ent->len);
			count_entries -= 1;
		}
		ISC_LIST_UNLINK(ent_list, ent, link);
		free(ent);
	}
}

static void
destroy(void) {
	struct qrent *ent, *entnext;
	int rc;

	for (ent = ISC_LIST_HEAD(ent_list);
	     ent != NULL;
	     ent = entnext)
	{
		entnext = ISC_LIST_NEXT(ent, link);
		JHSD(rc, urstate, &ent->key, ent->len);
		ISC_LIST_UNLINK(ent_list, ent, link);
		free(ent);
	}
	JHSFA(rc, urstate);
}

static int
process_msg(ncap_t ncap, ncap_msg_ct ncapmsg) {
	char qname[NS_MAXDNAME];
	struct qrval **val;
	int res = 0;
	int mlen, tlen, suppress, dofree = 0;
	ns_msg msg;
	ns_rr rr;
	struct qrent *ent;
	size_t rem;


	if (ns_initparse(ncapmsg->payload, ncapmsg->paylen, &msg) < 0)
		return (0);
	if (ns_msg_count(msg, ns_s_qd) != 1)
		return (0);
	if (ns_parserr(&msg, ns_s_qd, 0, &rr))
		return (0);
	
	count_messages_proc += 1;
	if (verbose && (count_messages_proc & 32767) == 0)
		fprintf(stderr, "mod_urstate] "
			"messages=%" PRIu64
			" entries=%" PRIu64
			" dupq=%" PRIu64
			" dupr=%" PRIu64
			" out=%" PRIu64
			"\n",
			count_messages_proc,
			count_entries,
			count_dup_query,
			count_dup_response,
			count_messages_out
		);

	mlen = tlen = dns_makename(ns_rr_name(rr), qname);
	rem = mlen & 63;
	if (rem != 0)
		mlen += (64 - rem);

	ent = calloc(1, sizeof(*ent) + mlen);
	assert(ent != NULL);
	ent->len = sizeof(struct qrkey) + tlen;

	if (!clear_id)
	ent->key.qid = ns_msg_id(msg);
	ent->key.qtype = ns_rr_type(rr);
	ent->key.qclass = ns_rr_class(rr);

	ent->key.ip[0] = ncapmsg->npu.ip4.src.s_addr;
	ent->key.ip[1] = ncapmsg->npu.ip4.dst.s_addr;
	if (!clear_ports) {
		ent->key.port[0] = ncapmsg->tpu.udp.sport;
	ent->key.port[1] = ncapmsg->tpu.udp.dport;
	}

	if (ent->key.ip[0] > ent->key.ip[1])
		SWAP(ent->key.ip[0], ent->key.ip[1]);
	if (ent->key.port[0] > ent->key.port[1])
		SWAP(ent->key.port[0], ent->key.port[1]);
	
	suppress = ((unsigned) ctx->now.tv_sec - start_time) < suppress_duration;
	memcpy(ent->key.qname, qname, tlen);

	JHSG(val, urstate, &ent->key, ent->len);
	if (val == NULL) {
		/* key does not exist */
		JHSI(val, urstate, &ent->key, ent->len);
		*val = calloc(1, sizeof(struct qrval));
		assert(*val != NULL);

		count_entries += 1;
		ISC_LINK_INIT(ent, link);
		ISC_LIST_APPEND(ent_list, ent, link);
	} else {
		/* key exists, we don't need this copy */
		dofree = 1;
	}

	if (ns_msg_getflag(msg, ns_f_qr)) {
		/* response */
		if ((*val)->msgr == NULL) {
			(*val)->count -= 1;
			(*val)->msgr = dup_msg((struct ncap_msg *) ncapmsg);
		} else if (same_msg((*val)->msgr, ncapmsg)) {
			count_dup_response += 1;
		} else {
			(*val)->count -= 1;
			if ((*val)->count < 0 && !suppress) {
				res = 1;
				/* output the two messages */
				if (NCAPTOOL_SINK_ENABLED(&urf)) {
					dump_msg(ncap, (*val)->msgr);
					dump_msg(ncap, ncapmsg);
				}
			}
		}

	} else {
		/* query */
		if ((*val)->msgq == NULL) {
			(*val)->count += 1;
			(*val)->msgq = dup_msg((struct ncap_msg *) ncapmsg);
		} else if (same_msg((*val)->msgq, ncapmsg)) {
			count_dup_query += 1;
		} else {
			(*val)->count += 1;
		}
	}

	if (dofree)
		free(ent);

	if (suppress)
		res = 0;

	if (pass_through)
		res = 1;
	return (res);
}

static void
dump_msg(ncap_t ncap, ncap_msg_ct msg) {
	FILE *dump = urf.fp;
	if (count_messages_out == 0)
		NCAP_OK(ncap, fwrite(ncap, NULL, dump),
			ctx->disablesink(&urf, ctx));
	if (NCAPTOOL_SINK_ENABLED(&urf))
		NCAP_OK(ncap, fwrite(ncap, msg, dump),
			ctx->disablesink(&urf, ctx));
	if (NCAPTOOL_SINK_ENABLED(&urf) && ctx->flush)
		fflush(dump);
	if (ferror(dump))
		ctx->disablesink(&urf, ctx);
	count_messages_out += 1;
}

static struct ncap_msg *
dup_msg(struct ncap_msg *orig) {
	struct ncap_msg *new;
	new = malloc(sizeof *new);
	assert(new != NULL);
	memcpy(new, orig, sizeof *new);
	new->payload = malloc(orig->paylen);
	assert(new->payload != NULL);
	memcpy((u_char *) new->payload, orig->payload, orig->paylen);
	return (new);
}

static void
free_msg(struct ncap_msg *msg) {
	if (msg != NULL) {
		free((u_char *) msg->payload);
		free(msg);
	}
}

static int
same_msg(ncap_msg_ct m1, ncap_msg_ct m2) {
	if (m1->paylen == m2->paylen)
		return (memcmp(m1->payload, m2->payload, m1->paylen) == 0);
	return (0);
}

static int
dns_makename(const char *pname, char *tname) {
	const char *tp1;
	char *tp2;
	int i;

	for (tp1 = pname, tp2 = tname, i = 0;
	     *tp1 != '\0';
	     tp1++, tp2++, i++)
	{
		int ch = *tp1;

		*tp2 = isascii(ch) && isupper(ch) ? tolower(ch) : ch;
	}
	*tp2 = '\0';
	return (i + 1);
}

static void
usage(const char *msg) {
	const char *name = ncaptool_msgmod_ctx.name;
	fprintf(stderr, "%s] usage error (%s)\n", name, msg);
	fprintf(stderr, "%s] options are:\n", name);
	fprintf(stderr, "%s]\t-r thresh   response threshold (default: %u)\n",
		name, RESPONSE_THRESHOLD);
	fprintf(stderr, "%s]\t-t secs     initial response suppression duration (default: %u)\n",
		name, SUPPRESS_DURATION);
	fprintf(stderr, "%s]\t-L limit	  maximum entries to allocate (default: %u)\n",
		name, ENTRIES_LIMIT);
	fprintf(stderr, "%s]\t-D          neglect qid matching (default: no)\n", name);
	fprintf(stderr, "%s]\t-P          neglect port pair matching (default: no)\n", name);
	fprintf(stderr, "%s]\t-v          verbose mode (default: no)\n", name);
}
